package org.netkernel.lang.trl.endpoint;

import java.io.ByteArrayInputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.netkernel.layer0.nkf.INKFAsyncRequestHandle;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.util.RequestBuilder;
import org.netkernel.layer0.util.XMLUtils;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.w3c.dom.Document;

public class TRLRuntime extends StandardAccessorImpl
{
	private static Pattern mCommentLinePattern=Pattern.compile("^#.*?(\\n|\\r)+",Pattern.DOTALL|Pattern.MULTILINE);
	private static Pattern mRequestPattern=Pattern.compile("(\\$[tea]{0,3}\\{)(.*?)\\}",Pattern.DOTALL|Pattern.MULTILINE);
	private static Pattern mDeclarativeRequestPattern=Pattern.compile(".*?<request>.*?",Pattern.DOTALL|Pattern.MULTILINE);
	
	public TRLRuntime()
	{	this.declareThreadSafe();		
	}

	@Override
	public void onSource(INKFRequestContext context) throws Exception
	{	String t=context.source("arg:template", String.class);
		INKFResponse resp=context.createResponseFrom(recurse2(t, context, context.getThisRequest().argumentExists("tolerant")));
		String mimetype = context.getThisRequest().getArgumentValue("mimetype");
		if (mimetype!=null)
		{	resp.setMimeType(mimetype);
		}
		else
		{	resp.setMimeType("text/plain");
		}
	}
	
	
	private static String recurse2(String aSource, INKFRequestContext aContext, boolean aTolerant) throws Exception
	{
		StringBuilder sb=stripComments2(aSource);
		
		Matcher m=mRequestPattern.matcher(sb);
		int start=0;
		List<ReplacementOp> rops=new ArrayList<ReplacementOp>();
		while(m.find(start))
		{	String head=m.group(1);
			String request=m.group(2);
			start=m.end();
			ReplacementOp rop=new ReplacementOp(m.start(), start, request, head);
			rops.add(rop);
		}
		
		for (ReplacementOp rop: rops)
		{	rop.doAsync(aContext);
		}
		
		int offset=0;
		for (ReplacementOp rop: rops)
		{	offset+=rop.processReplacement(sb,offset,aContext,aTolerant);
		}
		
		return sb.toString();
	}
	
	
	private static class ReplacementOp
	{
		private final int mStart;
		private final int mEnd;
		private final String mRequest;
		private final String mHeader;
		private INKFAsyncRequestHandle mAsyncHandle;
		
		public ReplacementOp(int aStart, int aEnd, String aRequest, String aHeader)
		{	mStart=aStart;
			mEnd=aEnd;
			mRequest=aRequest;
			mHeader=aHeader;
		}
		
		public void doAsync(INKFRequestContext aContext) throws Exception
		{	if (mHeader.contains("a"))
			{	INKFRequest req=TRLRuntime.createRequestFor(mRequest,aContext);
				mAsyncHandle=aContext.issueAsyncRequest(req);
			}
		}
		
		public int processReplacement(StringBuilder aBuffer, int aOffset, INKFRequestContext aContext, boolean aTolerant) throws Exception
		{
			String rep;
			try
			{	if (mAsyncHandle!=null)
				{	rep=(String)mAsyncHandle.join();
				}
				else
				{	INKFRequest req=TRLRuntime.createRequestFor(mRequest,aContext);
					rep=(String)aContext.issueRequest(req);
				}
			}
			catch (NKFException e)
			{	if (aTolerant)
				{	rep="{TLR-ERROR}"+e.getDeepestMessage()+"{/TLR-ERROR}";
				}
				else
				{	throw e;
				}
			}
			
			if(!mHeader.contains("t"))
			{	rep=recurse2(rep, aContext, aTolerant);
			}
			
			if (mHeader.contains("e"))
			{	rep=HTMLEncode(rep);				
			}
			
			aBuffer.replace(aOffset+mStart, aOffset+mEnd, rep);
			int subLength=mEnd-mStart;
			return rep.length()-subLength;
		}
		
	}
	
	private static INKFRequest createRequestFor(String aRequest, INKFRequestContext aContext) throws Exception
	{	Matcher m=mDeclarativeRequestPattern.matcher(aRequest);
		INKFRequest result;
		if(m.matches())
		{	Document d=XMLUtils.parse(new ByteArrayInputStream(aRequest.getBytes()));
			RequestBuilder b=new RequestBuilder(d, aContext.getKernelContext().getKernel().getLogger());
			result=b.buildRequest(aContext, null, null);
		}
		else
		{	result=aContext.createRequest(aRequest);
		}
		result.setRepresentationClass(String.class);
		return result;
	}
	
	private static StringBuilder stripComments2(String aSource) throws Exception
	{
		Matcher m=mCommentLinePattern.matcher(aSource);
		StringBuffer sb=new StringBuffer(aSource.length());	//Use StringBuffer as Matcher requires this and not Builder
		while(m.find())
		{	m.appendReplacement(sb, "");
		}
		m.appendTail(sb);
		return new StringBuilder(sb);
	}
		
	public static String HTMLEncode(String s)
	{   if(s==null)
		{	return null;
		}
        StringBuilder sb = new StringBuilder();
        for (int i = 0; i < s.length(); i++)
        {   char c = s.charAt(i);
            if (c >= 'a' && c <= 'z' || c >= 'A' && c <= 'Z' || c >= '0' && c <= '9')
            {	sb.append(c);
            }
            else
            {	sb.append("&#"+(int)c+";");
            }
        }
        return sb.toString();
    }
}
